# THIS FILE IS GENERATED FROM TRAITS SETUP.PY
version = '5.1.2'
full_version = '5.1.2'
git_revision = 'e62cb0a15195e14f5aa916f941d9a93a488a04fd'
is_released = True

if not is_released:
    version = full_version
